from __future__ import annotations
from copy import deepcopy
from dataclasses import dataclass, MISSING, _MISSING_TYPE
from functools import cached_property
from typing import Any, Literal, Optional, Union
from osin.models.exp import ExpRun
from osin.types.pyobject_type import NoneType


EXP_INDEX_FIELD = "__exp__"
EXPNAME_INDEX_FIELD_TYPE = Literal["__exp__"]

# Representing a record of a data source
Record = ExpRun

# For a single-level index, AttrValue is an element of the index
# For a multi-level index, AttrValue is an item of an n-tuple, which is an element of the index.
# We limit AttrValue to be Python primitive types, so that there is always an injection for the attribute
AttrValue = Union[str, int, bool, None]
# path to an attribute in a record
Attr = tuple[str, ...]
IndexElement = tuple[AttrValue, ...]


@dataclass
class AttrGetter:
    # path to the attribute in the tree
    path: Attr
    values: Optional[dict[AttrValue, int]] = None

    def get_attribute_value(self, record: Record) -> AttrValue | _MISSING_TYPE:
        """Get attribute value. If the filtered values are specified, if the value is not in the filtered values, return MISSING"""
        value = self._get_value(record)
        if value is MISSING:
            return value
        if isinstance(value, list) and all(
            isinstance(x, (str, int, bool, NoneType)) for x in value
        ):
            # support parameters that are a list of enumerated options
            value = ",".join(str(x) for x in sorted(value))
            return value
        if isinstance(value, (str, int, bool, NoneType)):
            return value
        raise ValueError(f"Invalid value: {value}")

    def get_value(self, record: Record) -> AttrValue | float | _MISSING_TYPE:
        """Get the attribute value, but this attribute value is not used in the index so it can be a float"""
        value = self._get_value(record)
        if value is not MISSING:
            if not isinstance(value, (str, int, float, bool, NoneType)):
                raise ValueError(f"Invalid value: {value}")
        return value

    def _get_value(self, record: Record) -> Any | _MISSING_TYPE:
        if self.path[0] == EXP_INDEX_FIELD:
            if self.values is not None and record.exp_id not in self.values:
                return MISSING
            return record.exp.name

        if self.path[0] == "params":
            props: dict = record.params
        elif self.path[0] == "aggregated_primitive_outputs":
            props: dict = record.aggregated_primitive_outputs
        else:
            raise KeyError(f"Invalid path: {self.path}")

        for i in range(1, len(self.path) - 1):
            if self.path[i] not in props:
                return MISSING
            tmp = props[self.path[i]]
            if not isinstance(tmp, dict):
                return MISSING
            props = tmp

        value = props.get(self.path[-1], MISSING)
        try:
            if self.values is not None and value not in self.values:
                return MISSING
        except TypeError:
            # value is not hashable
            raise ValueError(f"Invalid value: {value}")
        return value

    def to_tuple(self):
        return self.path, list(self.values.keys()) if self.values is not None else None

    @classmethod
    def from_tuple(cls, obj):
        return cls(
            tuple(obj[0]),
            {v: i for i, v in enumerate(obj[1])} if obj[1] is not None else None,
        )


@dataclass
class Index:
    attr: Attr
    # mapping from values at this level to the values at the lower level
    # the element of the index can be constructed from traveling down the tree
    # combining each value at each level
    # NOTE: we accept that the lower level can be a combination of values obtained
    # from different attribute getters. Therefore, we use list[Index] instead of Optional[Index]
    # when the list is empty, it means you have reached the leaf of the index
    children: dict[AttrValue, list[Index]]

    def merge(self, other: Index) -> None:
        """Merge two indexes"""
        assert self.attr == other.attr
        for value, other_next_values in other.children.items():
            if value in self.children:
                our_next_values = self.children[value]
                if len(our_next_values) > 0:
                    # other element and our element must have the same type
                    # because it should not be okay to have two paths that one path is a subpath of the other
                    # in that case, the subpath should be removed
                    assert len(other_next_values) > 0
                    # merge two lists
                    path2other_next_values = {x.attr: x for x in other_next_values}
                    for our_next_value in our_next_values:
                        if our_next_value.attr in path2other_next_values:
                            our_next_value.merge(
                                path2other_next_values[our_next_value.attr]
                            )
                            del path2other_next_values[our_next_value.attr]
                    for other_next_value in path2other_next_values.values():
                        our_next_values.append(other_next_value)
                else:
                    assert len(other_next_values) == 0
                    # do nothing
            else:
                self.children[value] = other_next_values

    def trim_unobserved_combinations(
        self, attrs: list[AttrGetter], fully_observed_combinations: list[IndexElement]
    ) -> bool:
        if len(attrs) == 0:
            return True

        if attrs[0].path != self.attr:
            # need to go deeper
            if all(element is None for element in self.children.values()):
                return False

            success = False
            for next_values in self.children.values():
                for element in next_values:
                    success = success or element.trim_unobserved_combinations(
                        attrs, fully_observed_combinations
                    )
            return success

        # we are at the right level
        value2subobservations = {}
        for combination in fully_observed_combinations:
            value = combination[0]
            if value not in value2subobservations:
                value2subobservations[value] = []
            value2subobservations[value].append(combination[1:])

            if value not in self.children:
                raise ValueError("Observe an attribute value that is not in the index")

        attr_keys: Optional[set[Attr]] = None

        for value, next_values in list(self.children.items()):
            if value not in value2subobservations:
                # this value is not observed, so we remove it from the index, but only the part of the index that contains attrs[1:]
                if attr_keys is None:
                    attr_keys = {attr.path for attr in attrs[1:]}
                next_values = [x for x in next_values if x.attr not in attr_keys]
                if len(next_values) == 0:
                    del self.children[value]
                else:
                    for next_level_value in next_values:
                        next_level_value.remove_attrs(attr_keys)
                    self.children[value] = next_values
            else:
                if len(next_values) == 0:
                    # we reach the lowest level, so the subobservations should be empty
                    if any(len(x) != 0 for x in value2subobservations[value]):
                        return False
                else:
                    success = False
                    for element in next_values:
                        success = success or element.trim_unobserved_combinations(
                            attrs[1:], value2subobservations[value]
                        )
                    if not success:
                        return False
        return True

    def remove_attrs(self, attrs: set[Attr]):
        for value, next_values in list(self.children.items()):
            if len(next_values) == 0:
                continue
            next_values = [x for x in next_values if x.attr not in attrs]
            if len(next_values) == 0:
                del self.children[value]
            else:
                for next_level_value in next_values:
                    next_level_value.remove_attrs(attrs)
                self.children[value] = next_values


@dataclass
class IndexSchema:
    attrs: list[AttrGetter]
    # mapping from attrgetter's index to the index of attrgetter's children
    index2children: list[list[int]]
    # list of attributes that are fully observed in the data source
    fully_observed_attrs: list[list[int]]

    def build_index(self, records: list[Record]) -> list[Index]:
        """Build an index from the list of records"""
        this = self._infer_attributes(records)
        paths = this.leaf_paths
        indices = [this._build_index_from_path(paths[0])]
        for path in paths[1:]:
            newindex = this._build_index_from_path(path)
            for index in indices:
                if index.attr == newindex.attr:
                    index.merge(newindex)
                    break
            else:
                indices.append(newindex)

        fully_observed_combinations = this._get_fully_observed_combinations(records)
        for attrs, combinations in fully_observed_combinations:
            trim_results = [
                index.trim_unobserved_combinations(attrs, combinations)
                for index in indices
            ]
            if not any(trim_results):
                raise ValueError(
                    "The order in the observed combinations is not consistent with the index"
                )

        return indices

    def get_index_element(self, record: Record) -> Optional[IndexElement]:
        """Get the index's element of a record"""
        paths = self.leaf_paths
        index_element = {}
        elements = []
        for path in paths:
            element = []
            for attrgetter in path:
                if attrgetter.values is None:
                    if attrgetter.path not in index_element:
                        index_element[attrgetter.path] = attrgetter.get_attribute_value(
                            record
                        )
                    if index_element[attrgetter.path] is MISSING:
                        element = None
                        break
                    element.append(index_element[attrgetter.path])
                else:
                    if (value := attrgetter.get_attribute_value(record)) is not MISSING:
                        element.append(value)
                    else:
                        element = None
                        break
            else:
                elements.append(tuple(element))

        if len(elements) == 0:
            return None
        elif len(elements) > 1:
            raise InvalidIndexError(
                f"This multi-level index is ambiguous. Found two index elements: {', '.join([str(element) for element in elements])} in the record {record.id}"
            )
        return tuple(elements[0])

    @staticmethod
    def from_tuple(obj):
        return IndexSchema(
            [AttrGetter.from_tuple(x) for x in obj[0]],
            obj[1],
            obj[2],
        )

    def to_tuple(self):
        return (
            [attrgetter.to_tuple() for attrgetter in self.attrs],
            self.index2children,
            self.fully_observed_attrs,
        )

    @cached_property
    def leaf_paths(self) -> list[list[AttrGetter]]:
        roots = [i for i, parents in enumerate(self.index2parents) if len(parents) == 0]
        paths = []
        for root in roots:
            paths.extend(self._leaf_paths_recur(root))

        return paths

    @cached_property
    def index2parents(self):
        index2parents_: list[list[int]] = [[] for i in range(len(self.attrs))]
        for i, children in enumerate(self.index2children):
            for ci in children:
                index2parents_[ci].append(i)
        return index2parents_

    def _infer_attributes(self, records: list[Record]) -> IndexSchema:
        """Infer attributes from the list of records"""
        this = deepcopy(self)
        for attrgetter in this.attrs:
            if attrgetter.values is None:
                values = {}
                for record in records:
                    value = attrgetter.get_attribute_value(record)
                    if value is not MISSING and value not in values:
                        values[value] = len(values)
                attrgetter.values = values

        return this

    def _get_fully_observed_combinations(
        self, records: list[Record]
    ) -> list[tuple[list[AttrGetter], list[IndexElement]]]:
        observed_combinations = []
        for observed_attr_idxs in self.fully_observed_attrs:
            observed_attrs = [self.attrs[i] for i in observed_attr_idxs]
            combinations = set()
            for record in records:
                elem = tuple(
                    attr.get_attribute_value(record) for attr in observed_attrs
                )
                if all(v is not MISSING for v in elem):
                    combinations.add(elem)
            observed_combinations.append((observed_attrs, combinations))
        return observed_combinations

    def _build_index_from_path(self, path: list[AttrGetter]) -> Index:
        if len(path) == 0:
            raise Exception("Path should not be empty. This is a bug.")

        attrgetter = path[0]
        assert attrgetter.values is not None

        if len(path) == 1:
            return Index(path[0].path, {value: [] for value in attrgetter.values})

        remaining_path = path[1:]
        return Index(
            path[0].path,
            {
                value: [self._build_index_from_path(remaining_path)]
                for value in attrgetter.values
            },
        )

    def _leaf_paths_recur(self, attrgetter_index: int) -> list[list[AttrGetter]]:
        attrgetter = self.attrs[attrgetter_index]
        if len(self.index2children[attrgetter_index]) == 0:
            return [[attrgetter]]

        paths = []
        for i in self.index2children[attrgetter_index]:
            for path in self._leaf_paths_recur(i):
                paths.append([attrgetter] + path)
        return paths


class InvalidIndexError(Exception):
    """Raised when the index is invalid."""

    pass



from osin.models.base import db, init_db
from osin.models.exp import Exp, ExpRun
from osin.models.report import ExpReport, Report
from osin.models.exp_data import ExpRunData, Record, ExampleData
from osin.models.views import ExpRunView

all_tables = [Exp, ExpRun, Report, ExpRunView, ExpReport]
__all__ = [
    "db",
    "init_db",
    "Exp",
    "ExpRun",
    "Report",
    "ExpRunView",
    "ExpReport",
    "ExpRunData",
    "Record",
    "ExampleData",
]



from datetime import datetime
from flask import jsonify, request
from gena import generate_api
from peewee import DoesNotExist, fn
from osin.misc import get_extension, identity
from osin.models.exp import Exp, ExpRun
from werkzeug.exceptions import BadRequest, NotFound
from osin.repository import OsinRepository
from werkzeug.utils import secure_filename

exp_bp = generate_api(Exp)
exprun_bp = generate_api(
    ExpRun,
    deserializers={
        "params": identity,
        # seems that with timezone, peewee cannot parse back to datetime
        # "finished_time": lambda x: parse(x).replace(tzinfo=None),
        # "created_time": lambda x: parse(x).replace(tzinfo=None),
        "aggregated_primitive_outputs": identity,
    },
)


@exprun_bp.route(f"/{exprun_bp.name}/activity", methods=["GET"])
def run_activity():
    if "since" in request.args:
        try:
            since = datetime.utcfromtimestamp(int(request.args.get("since", 0)) / 1000)
        except:
            raise BadRequest("since must be milliseconds since epoch")
    else:
        since = None
    day = fn.Strftime("%Y-%m-%d", ExpRun.created_time).alias("day")
    query = ExpRun.select(
        fn.Count(ExpRun.id).alias("n_runs"),
        day,
    )

    if since is not None:
        query = query.where(ExpRun.created_time > since)

    query = query.group_by(day)

    return jsonify([{"count": r.n_runs, "date": r.day} for r in query])


@exprun_bp.route(f"/{exprun_bp.name}/<id>/data", methods=["GET"])
def fetch_exp_run_data(id: int):
    """Fetch the experiment run data"""
    try:
        exp_run: ExpRun = ExpRun.get_by_id(id)
    except DoesNotExist:
        raise NotFound(f"ExpRun with id {id} does not exist")

    osin = OsinRepository.get_instance()
    format = osin.get_exp_run_data_format(exp_run.exp, exp_run)
    h5file = osin.get_exp_run_data_file(exp_run.exp, exp_run)

    limit = request.args.get("limit", "50")
    if not limit.isdigit():
        raise BadRequest("limit must be an integer")
    limit = int(limit)

    offset = request.args.get("offset", "0")
    if not offset.isdigit():
        raise BadRequest("offset must be an integer")
    offset = int(offset)

    if "fields" in request.args:
        fields = {}
        for field in request.args["fields"].split(","):
            parts = field.split(".")
            if len(parts) == 0 or len(parts) > 2:
                raise BadRequest(f"Invalid field: {field}")

            if parts[0] not in ["aggregated", "individual"]:
                raise BadRequest(f"Invalid field {field}")
            if len(parts) == 1:
                fields[parts[0]] = {"primitive", "complex"}
            else:
                if parts[1] not in {"primitive", "complex"}:
                    raise BadRequest(f"Invalid field {field}")
                fields.setdefault(parts[0], set()).add(parts[1])
    else:
        fields = None

    if "sorted_by" in request.args:
        sorted_by = request.args["sorted_by"].replace(".", "/")
        if sorted_by.startswith("-"):
            sorted_by = sorted_by[1:]
            sorted_order = "descending"
        else:
            sorted_order = "ascending"
    else:
        sorted_by = None
        sorted_order = "ascending"

    try:
        exp_run_data, n_examples = format.load_exp_run_data(
            h5file, fields, limit, offset, sorted_by, sorted_order
        )
    except KeyError:
        raise BadRequest(f"The key `{sorted_by}` does not exist to sort by")

    out = exp_run_data.to_dict()
    out["n_examples"] = n_examples
    return jsonify(out)


@exprun_bp.route(f"/{exprun_bp.name}/<id>/upload", methods=["POST"])
def upload_exp_run_data(id: int):
    try:
        exp_run: ExpRun = ExpRun.get_by_id(id)
    except DoesNotExist:
        raise NotFound(f"ExpRun with id {id} does not exist")

    files = {}
    for file_id, file in request.files.items():
        print(file, file and file.filename is not None, get_extension(file.filename))
        if (
            file
            and file.filename is not None
            and ("." + (get_extension(file.filename) or ""))
            in OsinRepository.ALLOWED_EXTENSIONS
        ):
            files[file_id] = file

    if len(files) == 0:
        raise BadRequest(f"No files provided")

    osin = OsinRepository.get_instance()
    rundir = osin.get_exp_run_dir(exp_run.exp, exp_run)
    if not rundir.exists():
        rundir.mkdir(parents=True)

    for file_id, file in files.items():
        filename = secure_filename(file.filename)
        file.save(str(rundir / filename))

    (rundir / "_SUCCESS").touch()

    return jsonify({"status": "success"})



import os
from pathlib import Path

import click
from loguru import logger
from peewee import fn
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.wsgi import WSGIContainer
from osin.models import db as dbconn, init_db, all_tables
from osin.repository import OsinRepository


@click.command()
@click.option("-d", "--data", required=True, help="data directory of osin")
def init(data):
    """Init database"""
    init_db(OsinRepository.get_instance(data).get_db_file())
    dbconn.create_tables(all_tables, safe=True)


@click.command()
@click.option("-d", "--data", required=True, help="data directory of osin")
@click.option("--wsgi", is_flag=True, help="Whether to use wsgi server")
@click.option("-p", "--port", default=5524, help="Listening port")
@click.option(
    "--certfile", default=None, help="Path to the certificate signing request"
)
@click.option("--keyfile", default=None, help="Path to the key file")
def start(
    data: str,
    wsgi: bool,
    port: int,
    certfile: str,
    keyfile: str,
):
    init_db(OsinRepository.get_instance(data).get_db_file())

    if certfile is None or keyfile is None:
        ssl_options = None
    else:
        ssl_options = {"certfile": certfile, "keyfile": keyfile}
        assert not wsgi

    from osin.app import app

    if wsgi:
        app.run(host="0.0.0.0", port=port)
    else:
        logger.info("Start server in non-wsgi mode")
        http_server = HTTPServer(WSGIContainer(app), ssl_options=ssl_options)
        http_server.listen(port)
        IOLoop.instance().start()


@click.group()
def cli():
    pass


cli.add_command(init)
cli.add_command(start)


if __name__ == "__main__":
    cli()



from pathlib import Path
from typing import Union

from peewee import SqliteDatabase, Model, Field


# TODO: consider moving to APSWDatabase
db = SqliteDatabase(None)
is_inited = False


def init_db(dbfile: Union[str, Path]):
    """Initialize database"""
    global db, is_inited
    if is_inited:
        return

    db.init(str(dbfile), pragmas={"foreign_keys": 1})
    is_inited = True


class BaseModel(Model):
    class Meta:
        database = db


class ClassField(Field):
    field_type = "BLOB"

    def __init__(self, cls, **kwargs):
        super().__init__(**kwargs)
        self.db_value = getattr(cls, "db_value")
        self.python_value = getattr(cls, "python_value")


class BlobField(Field):
    field_type = "BLOB"

    def __init__(self, serialize, deserialize, **kwargs):
        super().__init__(**kwargs)
        self.db_value = serialize
        self.python_value = deserialize



from osin.models.report.dbmodel import (
    ExpReport,
    Report,
    ReportArgs,
    ReportDisplayPosition,
    ReportType,
)
from osin.models.report.base_report import BaseReport
from osin.models.report.index_schema import (
    EXP_INDEX_FIELD,
    EXPNAME_INDEX_FIELD_TYPE,
    IndexSchema,
    AttrGetter,
)

__all__ = [
    "ExpReport",
    "Report",
    "ReportArgs",
    "EXP_INDEX_FIELD",
    "EXPNAME_INDEX_FIELD_TYPE",
    "ReportDisplayPosition",
    "ReportType",
    "BaseReport",
    "IndexSchema",
    "AttrGetter",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Union, Mapping
from osin.types.pyobject_type import PRIMITIVE_TYPES, Number, PyObjectType


PrimitiveValue = Union[str, int, float, bool, None]
NestedPrimitiveOutput = dict[str, Union[PrimitiveValue, "NestedPrimitiveOutput"]]


class InvalidValueType(Exception):
    def __init__(self, valid_type: str):
        super().__init__(self)
        self.valid_type = valid_type
        self.trace = []

    def add_trace(self, *parents: str):
        self.trace.extend(reversed(parents))
        return self

    def __str__(self):
        return f"Expect {self.valid_type} but receive value type: {'·'.join(reversed(self.trace))}"


@dataclass
class NestedPrimitiveOutputSchema:
    schema: Dict[str, Union[PyObjectType, NestedPrimitiveOutputSchema]]

    @staticmethod
    def from_tuple(object):
        schema = {}
        for prop, prop_type in object[0].items():
            if isinstance(prop_type[0], dict):
                schema[prop] = NestedPrimitiveOutputSchema.from_tuple(prop_type)
            else:
                schema[prop] = PyObjectType.from_tuple(prop_type)
        return NestedPrimitiveOutputSchema(schema)

    @staticmethod
    def infer_from_data(
        data: Mapping[str, Any], use_number_type: bool = True
    ) -> NestedPrimitiveOutputSchema:
        schema = {}
        for key, value in data.items():
            if isinstance(value, dict):
                schema[key] = NestedPrimitiveOutputSchema.infer_from_data(
                    value, use_number_type
                )
            elif isinstance(value, str):
                schema[key] = PRIMITIVE_TYPES[str]
            elif isinstance(value, int):
                if use_number_type:
                    schema[key] = PRIMITIVE_TYPES[Number]
                else:
                    schema[key] = PRIMITIVE_TYPES[int]
            elif isinstance(value, float):
                if use_number_type:
                    schema[key] = PRIMITIVE_TYPES[Number]
                else:
                    schema[key] = PRIMITIVE_TYPES[float]
            elif isinstance(value, bool):
                schema[key] = PRIMITIVE_TYPES[bool]
            else:
                raise ValueError("{} is not a primitive type".format(value))

        return NestedPrimitiveOutputSchema(schema=schema)

    def does_data_match(self, data: Mapping[str, Any]) -> bool:
        if set(self.schema.keys()) != set(data.keys()):
            return False

        for prop, prop_schema in self.schema.items():
            value = data[prop]
            if isinstance(prop_schema, NestedPrimitiveOutputSchema):
                if not prop_schema.does_data_match(value):
                    return False
            else:
                if not prop_schema.is_instance(value):
                    return False

        return True

    def __repr__(self) -> str:
        output = []
        for prop, prop_schema in self.schema.items():
            if isinstance(prop_schema, NestedPrimitiveOutputSchema):
                output.append(f"{prop}:")
                for line in str(prop_schema).split("\n"):
                    output.append("    " + line)
            else:
                output.append(f"{prop}: {prop_schema}")

        return "\n".join(output)


def validate_primitive_data(data: NestedPrimitiveOutput):
    for key, value in data.items():
        if value is not None and not isinstance(value, (str, bool, int, float, dict)):
            raise InvalidValueType(
                "Nested Primitive Type (str, int, bool, float, None)"
            ).add_trace(key, str(type(value)))

        if isinstance(value, dict):
            try:
                validate_primitive_data(value)
            except InvalidValueType as e:
                raise e.add_trace(key)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from gena.custom_fields import DataClassField
from osin.models.base import BaseModel
from osin.models.exp import Exp
from osin.models.report.base_report import BaseReport
from peewee import CharField, ForeignKeyField, TextField


class ReportType(str, Enum):
    Table = "table"


@dataclass
class ReportArgs:
    type: ReportType
    value: BaseReport

    @staticmethod
    def from_tuple(obj: tuple):
        return ReportArgs(ReportType(obj[0]), BaseReport.from_tuple(obj[1]))

    def to_tuple(self):
        return (self.type.value, self.value.to_tuple())


@dataclass
class ReportDisplayPosition:
    # higher is closer to the top
    row_order: int
    # sum of col span & offset must be <= 24
    col_span: int
    col_offset: int


class Report(BaseModel):
    id: int
    name = CharField()
    description = TextField()
    args: ReportArgs = DataClassField(ReportArgs, null=True)  # type: ignore


class ExpReport(BaseModel):
    class Meta:
        indexes = ((("exp", "report"), True),)

    exp = ForeignKeyField(Exp, backref="reports", on_delete="CASCADE")
    exp_id: int
    report = ForeignKeyField(Report, backref="exps", on_delete="CASCADE")
    report_id: int

    position: ReportDisplayPosition = DataClassField(ReportDisplayPosition, null=True)  # type: ignore



from __future__ import annotations
from dataclasses import dataclass, is_dataclass
from typing import Any, Optional, Union, get_args, get_origin, get_type_hints

from osin.types.pyobject_type import PyObjectType, NoneType


DataClassInstance = Any


@dataclass
class ParamSchema:
    type: PyObjectType
    schema: dict[str, Union[PyObjectType, ParamSchema]]

    @staticmethod
    def from_tuple(object):
        type = PyObjectType.from_tuple(object[0])
        schema = {}
        for prop, prop_type in object[1].items():
            if isinstance(prop_type[1], dict):
                schema[prop] = ParamSchema.from_tuple(prop_type)
            else:
                schema[prop] = PyObjectType.from_tuple(prop_type)
        return ParamSchema(type, schema)

    @staticmethod
    def get_schema(param: DataClassInstance) -> ParamSchema:
        """Derive parameter types from a dataclass"""
        assert is_dataclass(
            param
        ), "Parameters must be an instance of a dataclass or a dataclass"
        param_cls = param if isinstance(param, type) else param.__class__
        type_hints = get_type_hints(param_cls)
        output = {}
        for name, hint in type_hints.items():
            if name in output:
                raise KeyError("Duplicate parameter name: {}".format(name))

            if is_dataclass(hint):
                output[name] = ParamSchema.get_schema(hint)
            else:
                origin = get_origin(hint)
                args = get_args(hint)

                if ParamSchema.is_optional_type(origin, args):
                    subtype = args[0] if args[0] != NoneType else args[1]
                    if is_dataclass(subtype):
                        schema = ParamSchema.get_schema(subtype)
                        schema.type = schema.type.make_optional()
                        output[name] = schema
                        continue
                output[name] = PyObjectType.from_type_hint(hint)
        return ParamSchema(PyObjectType.from_type_hint(param_cls), output)

    @staticmethod
    def is_optional_type(origin: Optional[type], args: tuple[type, ...]):
        return origin is Union and len(args) == 2 and NoneType in args



import os
from gena import generate_app

from osin.controllers.exp import exp_bp, exprun_bp
from osin.controllers.report import report_bp, expreport_bp
from osin.controllers.views import exprunview_bp

app = generate_app(
    [
        exp_bp,
        exprun_bp,
        exprunview_bp,
        report_bp,
        expreport_bp,
    ],
    os.path.dirname(__file__),
    # log_sql_queries=False,
)

app.config["MAX_CONTENT_LENGTH"] = 100 * 1024 * 1024  # maximum upload of 100 MB



import os

from pathlib import Path


_ROOT_DIR = Path(os.path.abspath(__file__)).parent.parent
PACKAGE_DIR = str(Path(os.path.abspath(__file__)).parent)
FROM_SITEPACKAGES = _ROOT_DIR.name == "site-packages"

CACHE_SIZE = 10240

SETTINGS = {}



from dataclasses import dataclass
import math, numpy as np
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Literal,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from numpy import sort
from osin.models import ExpRunData, ExampleData, Record
from osin.types import NestedPrimitiveOutput, PyObject
from h5py import Group, File, Empty


class Hdf5Format:
    """An interface for storing experiment data"""

    def save_run_data(self, data: ExpRunData, outfile: Union[Path, str]):
        """Save the experiment run data to the file"""
        with File(outfile, "a") as f:
            self._update_nested_primitive_object(
                f.create_group("/aggregated/primitive", track_order=True),
                data.aggregated.primitive,
            )

            grp = f.create_group("/aggregated/complex", track_order=True)
            for key, value in data.aggregated.complex.items():
                self._validate_key(key)
                grp[key] = value.serialize_hdf5()
                grp.attrs[key] = value.get_classpath()

            ind_group = f.create_group("/individual", track_order=True)
            for (
                example_id,
                example,
            ) in data.individual.items():
                self._validate_key(example_id)
                ex_group = ind_group.create_group(example_id, track_order=True)
                ex_group.attrs["id"] = example.id
                ex_group.attrs["name"] = example.name

                self._update_nested_primitive_object(
                    ex_group.create_group("primitive", track_order=True),
                    example.data.primitive,
                )

                grp = ex_group.create_group(f"complex", track_order=True)
                for key, obj in example.data.complex.items():
                    self._validate_key(key)
                    grp[key] = obj.serialize_hdf5()
                    grp.attrs[key] = obj.get_classpath()

    def load_exp_run_data(
        self,
        infile: Union[Path, str],
        fields: Optional[Dict[str, Set[str]]] = None,
        limit: int = -1,
        offset: int = 0,
        sorted_by: Optional[str] = None,
        sorted_order: Literal["ascending", "descending"] = "ascending",
    ) -> Tuple[ExpRunData, int]:
        """Load experiment run data from a file.

        Args:
            infile: The file to load from
            fields: A dictionary of fields to load. If None, load all fields
            limit: The maximum number of individual records to load. If -1, load all records. Only apply to `individual` property
            offset: The offset to start loading from. Only apply to `individual` property
            sorted_by: The field to sort by. Only apply to `individual` property. Support either: id, name, or `data.<nested_field>` where `nested_field` is a field in the primitive or complex object
            sorted_order: The order to sort by. Only apply to `individual` property
        """
        if fields is None:
            fields = {
                "aggregated": {"primitive", "complex"},
                "individual": {"primitive", "complex"},
            }

        expdata = ExpRunData()
        with File(infile, "r") as f:
            if "aggregated" in fields:
                group = f["aggregated"]
                if "primitive" in fields["aggregated"]:
                    expdata.aggregated.primitive = self._load_nested_primitive_object(
                        group["primitive"]
                    )
                if "complex" in fields["aggregated"]:
                    for key, value in group["complex"].items():
                        pyobject_class = PyObject.from_classpath(group.attrs[key])
                        expdata.aggregated.complex[key] = pyobject_class.from_hdf5(
                            value[()]
                        )
            if "individual" in fields:
                if limit <= 0 and offset == 0 and sorted_by is None:
                    # select all without sorting
                    selected_examples = f["individual"].items()
                elif sorted_by is None:
                    if limit <= 0:
                        limit = math.inf  # type: ignore
                    # select some, no need to sort
                    selected_examples = []
                    for i, (key, ex_group) in enumerate(f["individual"].items()):
                        if i < offset:
                            continue
                        if i >= offset + limit:
                            break
                        selected_examples.append((key, ex_group))
                else:
                    if sorted_by.find("/") == -1:
                        sorted_by_first, sorted_by_remain = sorted_by, None
                        assert sorted_by_first in ["id", "name"], sorted_by_first
                    else:
                        sorted_by_first, sorted_by_remain = sorted_by.split("/", 1)
                        assert sorted_by_first == "data", sorted_by_first

                    # must sort
                    if limit <= 0:
                        limit = math.inf  # type: ignore
                    selected_examples = []
                    sorted_keys = {}
                    if sorted_by_remain is None:
                        for key, ex_group in f["individual"].items():
                            selected_examples.append((key, ex_group))
                            sorted_keys[key] = ex_group.attrs[sorted_by_first]
                    else:
                        for key, ex_group in f["individual"].items():
                            selected_examples.append((key, ex_group))
                            if sorted_by_remain not in ex_group:
                                raise KeyError(
                                    f"sort by key `data.{sorted_by_remain}` not found"
                                )
                            sorted_keys[key] = ex_group[sorted_by_remain][()]

                    selected_examples.sort(
                        key=lambda x: sorted_keys[x[0]],
                        reverse=sorted_order == "descending",
                    )
                    selected_examples = selected_examples[offset : offset + limit]

                for example_id, ex_group in selected_examples:
                    if example_id not in expdata.individual:
                        example = ExampleData(
                            id=ex_group.attrs["id"],
                            name=ex_group.attrs["name"],
                        )

                        if "primitive" in fields["individual"]:
                            example.data.primitive = self._load_nested_primitive_object(
                                ex_group["primitive"]
                            )
                        if "complex" in fields["individual"]:
                            for key, value in ex_group["complex"].items():
                                pyobject_class = PyObject.from_classpath(
                                    ex_group["complex"].attrs[key]
                                )
                                example.data.complex[key] = pyobject_class.from_hdf5(
                                    value[()]
                                )
                        expdata.individual[example_id] = example
                    else:
                        if "primitive" in fields["individual"]:
                            expdata.individual[
                                example_id
                            ].data.primitive = self._load_nested_primitive_object(
                                ex_group["primitive"]
                            )

                        if "complex" in fields["individual"]:
                            for key, value in ex_group["complex"].items():
                                pyobject_class = PyObject.from_classpath(
                                    ex_group.attrs[key]
                                )
                                expdata.individual[example_id].data.complex[
                                    key
                                ] = pyobject_class.from_hdf5(value[()])

            n_examples = len(f["individual"])
        return expdata, n_examples

    def _update_nested_primitive_object(
        self, group: Group, primitive_object: NestedPrimitiveOutput
    ):
        for key, value in primitive_object.items():
            self._validate_key(key)
            if isinstance(value, dict):
                subgroup = group.create_group(key, track_order=True)
                self._update_nested_primitive_object(subgroup, value)
            elif value is None:
                group[key] = Empty("f")
            else:
                group[key] = value

    def _load_nested_primitive_object(self, group: Group) -> NestedPrimitiveOutput:
        primitive_object = {}
        for key, value in group.items():
            if isinstance(value, Group):
                primitive_object[key] = self._load_nested_primitive_object(value)
            else:
                val = value[()]
                if isinstance(val, np.floating):
                    val = float(val)
                elif isinstance(val, np.integer):
                    val = int(val)
                elif isinstance(val, np.bool_):
                    val = bool(val)
                elif isinstance(val, bytes):
                    val = val.decode()
                elif isinstance(val, Empty):
                    val = None
                primitive_object[key] = val
        return primitive_object

    def _validate_key(self, key: str):
        if key.find("/") != -1:
            raise KeyError(f"Cannot have '/' in hdf5 group's item: {key}")



from pathlib import Path
from typing import Optional, Union

from osin.models.exp import Exp, ExpRun
from osin.formats import Hdf5Format
from osin.apis.remote_exp import RemoteExp, RemoteExpRun
from slugify import slugify


class OsinRepository:
    instance = None
    ALLOWED_EXTENSIONS = {".json", ".h5"}

    def __init__(self, osin_dir: Union[Path, str]):
        self.osin_dir = Path(osin_dir)
        self.osin_dir.mkdir(exist_ok=True, parents=True)

    @staticmethod
    def get_instance(osin_dir: Optional[Union[Path, str]] = None) -> "OsinRepository":
        if OsinRepository.instance is None:
            assert osin_dir is not None
            OsinRepository.instance = OsinRepository(osin_dir)
        elif osin_dir is not None:
            assert OsinRepository.instance.osin_dir == Path(osin_dir)
        return OsinRepository.instance

    def get_db_file(self) -> Path:
        return self.osin_dir / "osin.db"

    def get_exp_dir(
        self, exp: Union[Exp, RemoteExp], create_if_missing: bool = True
    ) -> Path:
        expdir = self.osin_dir / f"{slugify(exp.name)}-{exp.id}" / str(exp.version)
        if create_if_missing:
            expdir.mkdir(parents=True, exist_ok=True)
        return expdir

    def get_exp_run_dir(
        self, exp: Union[Exp, RemoteExp], exp_run: Union[ExpRun, RemoteExpRun]
    ) -> Path:
        return self.get_exp_dir(exp) / f"run_{exp_run.id:03d}"

    def get_exp_run_params_file(
        self, exp: Union[Exp, RemoteExp], exp_run: Union[ExpRun, RemoteExpRun]
    ) -> Path:
        return self.get_exp_run_dir(exp, exp_run) / "params.json"

    def get_exp_run_metadata_file(
        self, exp: Union[Exp, RemoteExp], exp_run: Union[ExpRun, RemoteExpRun]
    ) -> Path:
        return self.get_exp_run_dir(exp, exp_run) / "metadata.json"

    def get_exp_run_data_file(
        self, exp: Union[Exp, RemoteExp], exp_run: Union[ExpRun, RemoteExpRun]
    ) -> Path:
        return self.get_exp_run_dir(exp, exp_run) / "data.h5"

    def get_exp_run_data_format(
        self, exp: Union[Exp, RemoteExp], exp_run: Union[ExpRun, RemoteExpRun]
    ) -> Hdf5Format:
        return Hdf5Format()

    def get_exp_run_success_file(
        self, exp: Union[Exp, RemoteExp], exp_run: Union[ExpRun, RemoteExpRun]
    ) -> Path:
        return self.get_exp_run_dir(exp, exp_run) / "_SUCCESS"



from pathlib import Path
import sqlite3
import sys
from typing import Optional

import orjson


def identity(x):
    return x


def get_caller_python_script():
    """Determine the python script that starts the python program"""
    return sys.argv[0]


def json_dumps(obj):
    return orjson.dumps(obj, default=_orjson_default).decode()


def orjson_dumps(obj, **kwargs):
    return orjson.dumps(obj, default=_orjson_default, **kwargs)


def assert_not_null(obj):
    if obj is None:
        raise ValueError("expect a not-null object")
    return obj


def _orjson_default(obj):
    if isinstance(obj, Path):
        return str(obj)
    raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")


class Directory:
    def __init__(self, root: Path, dirname: str = "directory"):
        self.root = root
        self.root.mkdir(exist_ok=True, parents=True)
        self.dbfile = root / "fs.db"
        self.dirname = dirname

        need_init = not self.dbfile.exists()
        self.db = sqlite3.connect(str(self.dbfile))
        if need_init:
            with self.db:
                self.db.execute("CREATE TABLE files(path, diskpath, key)")

    def create_directory(self, relpath: str, key: dict, save_key: bool = False) -> Path:
        """Create a directory at a virtual relpath with key if not exists. Otherwise, return
        the existing directory"""
        ser_key = orjson_dumps(key)
        if save_key:
            # serialize here to make sure we can dump it first otherwise, we may have create
            # the directory but fail to dump the key
            friendly_ser_key = orjson_dumps(key, option=orjson.OPT_INDENT_2)
        else:
            friendly_ser_key = b""

        with self.db:
            record = self.db.execute(
                "SELECT path, diskpath, key FROM files WHERE path = ? AND key = ?",
                (relpath, ser_key),
            ).fetchone()

            if record is None:
                last_id = self.db.execute("SELECT MAX(rowid) FROM files").fetchone()[0]
                if last_id is None:
                    last_id = 0
                dirname = f"{self.dirname}_{last_id + 1:03d}"
                self.db.execute(
                    "INSERT INTO files VALUES (?, ?, ?)", (relpath, dirname, ser_key)
                )
                dpath = self.root / dirname
                dpath.mkdir(exist_ok=False, parents=True)
                if save_key:
                    (dpath / "_KEY").write_bytes(friendly_ser_key)

                return dpath
            return self.root / record[1]


def get_extension(filename: str) -> Optional[str]:
    """Return the extension of a filename without the dot"""
    lst = filename.rsplit(".", 1)
    if len(lst) == 1:
        return None
    return lst[1].lower()



import atexit
import shutil
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

import orjson
from loguru import logger
from osin.apis.osin import Osin
from osin.apis.remote_exp import RemoteExp, RemoteExpRun
from osin.misc import get_caller_python_script, orjson_dumps
from osin.models.base import init_db
from osin.models.exp import Exp, ExpRun, NestedPrimitiveOutput, RunMetadata
from osin.models.exp_data import ExampleData, Record
from osin.params_helper import DataClassInstance, param_as_dict
from osin.repository import OsinRepository
from osin.types import NestedPrimitiveOutputSchema, ParamSchema, PyObject
from osin.types.primitive_type import validate_primitive_data


class LocalOsin(Osin):
    def __init__(self, osin_dir: Union[Path, str]):
        super().__init__(osin_dir)
        init_db(self.osin_keeper.get_db_file())

    def _find_latest_exp(self, name: str) -> Optional[Exp]:
        exps = (
            Exp.select().where(Exp.name == name).order_by(Exp.version.desc()).limit(1)  # type: ignore
        )
        if len(exps) == 0:
            return None
        else:
            return exps[0]

    def _create_exp(self, exp: Exp) -> Exp:
        if exp.description is None or exp.params is None:
            raise ValueError(
                "Cannot create a new experiment without description and params"
            )
        exp.save(force_insert=True)
        return exp

    def _update_exp(self, exp_id: int, exp: Exp, fields: List[str]):
        Exp.update(**{field: getattr(exp, field) for field in fields}).where(Exp.id == exp_id).execute()  # type: ignore

    def _create_exprun(self, exprun: ExpRun) -> ExpRun:
        exprun.save(force_insert=True)
        return exprun

    def _update_exprun(self, exprun_id: int, exprun: ExpRun, fields: List[str]):
        ExpRun.update(**{field: getattr(exprun, field) for field in fields}).where(ExpRun.id == exprun_id).execute()  # type: ignore

    def _upload_exprun(self, exprun: RemoteExpRun):
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, List, Type, TypeVar

import numpy as np
import orjson
from osin.misc import orjson_dumps
from osin.types.primitive_type import NestedPrimitiveOutput
from osin.types.pyobject_type import PyObjectType

T = TypeVar("T", np.ndarray, bytes)


class PyObject(ABC, Generic[T]):
    def get_classpath(self) -> str:
        return PyObjectType.from_type_hint(self.__class__).path

    @staticmethod
    def from_classpath(classpath: str) -> Type[PyObject]:
        # we know that variants of pyobject must be member of this module
        return globals()[classpath.split(".")[-1]]

    @abstractmethod
    def serialize_hdf5(self) -> T:
        pass

    @staticmethod
    @abstractmethod
    def from_hdf5(value: T) -> PyObject:
        pass

    @abstractmethod
    def to_dict(self) -> dict:
        pass


@dataclass
class OTable(PyObject[bytes]):
    rows: List[NestedPrimitiveOutput]

    def serialize_hdf5(self) -> bytes:
        return orjson_dumps({"rows": self.rows})

    @staticmethod
    def from_hdf5(value: bytes) -> OTable:
        return OTable(**orjson.loads(value))

    def to_dict(self) -> dict:
        return {
            "type": "table",
            "rows": self.rows,
        }


@dataclass
class OImage(PyObject[np.ndarray]):
    object: np.ndarray

    def serialize_hdf5(self) -> np.ndarray:
        raise NotImplementedError()

    @staticmethod
    def from_hdf5(value: np.ndarray) -> OImage:
        raise NotImplementedError()

    def to_dict(self) -> Any:
        raise NotImplementedError()


@dataclass
class OAudio(PyObject[np.ndarray]):
    object: np.ndarray

    def serialize_hdf5(self) -> np.ndarray:
        raise NotImplementedError()

    @staticmethod
    def from_hdf5(value: np.ndarray) -> OAudio:
        raise NotImplementedError()

    def to_dict(self) -> dict:
        raise NotImplementedError()



from __future__ import annotations
from typing import Optional


from flask import jsonify, request
from gena import generate_api
from gena.api_generator import APIFuncs
from gena.deserializer import (
    deserialize_int,
    generate_deserializer,
    get_dataclass_deserializer,
    get_deserialize_list,
    get_deserializer_from_type,
)
from gena.serializer import get_peewee_serializer, get_dataclass_serializer
from osin.models.base import db
from osin.models.exp import Exp, ExpRun
from osin.models.report import (
    EXP_INDEX_FIELD,
    ExpReport,
    Report,
    ReportDisplayPosition,
)
from osin.models.report.index_schema import (
    AttrGetter,
    AttrValue,
    Index,
    InvalidIndexError,
)
from peewee import DoesNotExist
from werkzeug.exceptions import BadRequest, NotFound

expreport_bp = generate_api(ExpReport)
report_bp = generate_api(
    Report,
    skip_funcs={APIFuncs.create, APIFuncs.update},
    known_type_serializer={
        AttrGetter: lambda x: {
            "path": x.path,
            "values": list(x.values.keys()) if x.values is not None else None,
        }
    },
)

AttrGetter_values_deser = get_deserializer_from_type(Optional[list[AttrValue]], {})
report_deserializers = generate_deserializer(
    Report,
    known_type_deserializers={
        AttrGetter: get_dataclass_deserializer(
            AttrGetter,
            known_field_deserializers={
                "values": lambda x: {v: i for i, v in enumerate(resp)}
                if (resp := AttrGetter_values_deser(x)) is not None
                else None
            },
        )
    },
)
report_serializer = get_peewee_serializer(
    Report,
    known_type_serializer={
        AttrGetter: get_dataclass_serializer(
            AttrGetter,
            known_field_serializer={
                "values": lambda x: list(x.keys()) if x is not None else None
            },
        )
    },
)
exp_ids_deser = get_deserialize_list(deserialize_int)
pos_deser = get_dataclass_deserializer(
    ReportDisplayPosition, known_type_deserializers={}
)


@report_bp.route(f"/{report_bp.name}/get-attr-values", methods=["GET"])
def get_index_values():
    """Get possible values of an index"""
    exp_ids = request.args.get("exps")
    if exp_ids is None:
        raise BadRequest("Missing `exps` parameter")
    exp_ids = exp_ids.split(",")
    if not all(x.isdigit() for x in exp_ids):
        raise BadRequest("Invalid `exps` parameter")

    attr = request.args.get("attr")
    if attr is None:
        raise BadRequest("Missing `attr` parameter")

    if attr == EXP_INDEX_FIELD:
        exps = Exp.select(Exp.name).where(Exp.id.in_(exp_ids)).distinct()
        return jsonify({"items": [exp.name for exp in exps]})

    attrgetter = AttrGetter(path=tuple(attr.split(".")), values=None)
    runs = list(
        ExpRun.select(
            ExpRun.id, ExpRun.params, ExpRun.aggregated_primitive_outputs
        ).where(
            (ExpRun.exp.in_(exp_ids))
            & (ExpRun.is_deleted == False)
            & (ExpRun.is_successful == True)
        )
    )

    items = set()
    for run in runs:
        try:
            item = attrgetter.get_attribute_value(run)
        except KeyError:
            continue
        items.add(item)

    return jsonify({"items": list(items)})


@report_bp.route(f"/{report_bp.name}", methods=["POST"])
def create_report():
    posted_record = request.json

    if "exps" not in posted_record:
        raise BadRequest("Missing `exps` field")
    exp_ids = exp_ids_deser(posted_record["exps"])
    if len(exp_ids) == 0:
        raise BadRequest("Empty `exps` field")

    if "exp" not in posted_record:
        raise BadRequest("Missing `exp` field")
    main_exp_id = deserialize_int(posted_record["exp"])
    if main_exp_id not in exp_ids:
        raise BadRequest("Main exp must be in `exps` field")

    if "position" in posted_record:
        position = pos_deser(posted_record["position"])
    else:
        position = None

    raw_record = {}
    for name, field in Report._meta.fields.items():
        if name in posted_record and name != "id":
            try:
                raw_record[name] = report_deserializers[name](posted_record[name])
            except ValueError as e:
                raise ValueError(f"Field `{name}` {str(e)}")
    report = Report(**raw_record)

    with db:
        report.save()
        expreports = [{"exp_id": expid, "report_id": report.id} for expid in exp_ids]
        for r in expreports:
            if r["exp_id"] == main_exp_id:
                r["position"] = position
        ExpReport.insert_many(expreports).execute()

    return jsonify(report_serializer(report))


@report_bp.route(f"/{report_bp.name}/<id>", methods=["PUT"])
def update(id):
    try:
        record = Report.get_by_id(id)
    except DoesNotExist as e:
        raise NotFound(f"Record {id} does not exist")

    posted_record = request.get_json()
    if posted_record is None:
        raise BadRequest("Missing request body")

    if "exps" not in posted_record:
        raise BadRequest("Missing `exps` field")
    exp_ids = exp_ids_deser(posted_record["exps"])
    if len(exp_ids) == 0:
        raise BadRequest("Empty `exps` field")

    if "exp" not in posted_record:
        raise BadRequest("Missing `exp` field")
    main_exp_id = deserialize_int(posted_record["exp"])
    if main_exp_id not in exp_ids:
        raise BadRequest("Main exp must be in `exps` field")

    if "position" in posted_record:
        position = pos_deser(posted_record["position"])
    else:
        position = None

    for name, field in Report._meta.fields.items():
        if name in posted_record:
            try:
                value = report_deserializers[name](posted_record[name])
            except ValueError as e:
                raise ValueError(f"Field `{name}` {str(e)}")

            setattr(record, name, value)

    with db:
        record.save()
        ExpReport.delete().where(ExpReport.report_id == record.id).execute()
        expreports = [{"exp_id": expid, "report_id": record.id} for expid in exp_ids]
        for r in expreports:
            if r["exp_id"] == main_exp_id:
                r["position"] = position
        ExpReport.insert_many(expreports).execute()

    return jsonify(report_serializer(record))


@report_bp.route(f"/{report_bp.name}/<id>/data", methods=["GET"])
def get_report_data(id: int):
    """Get the report data"""

    # get the experiments
    report: Report = Report.get_by_id(id)
    exp_ids = [
        x.exp_id for x in ExpReport.select(ExpReport.exp).where(ExpReport.report == id)
    ]
    exps = {e.id: e for e in Exp.select(Exp.id, Exp.name).where(Exp.id.in_(exp_ids))}

    # gather runs of experiments
    runs = list(
        ExpRun.select(
            ExpRun.id, ExpRun.params, ExpRun.aggregated_primitive_outputs, ExpRun.exp
        ).where(
            (ExpRun.exp.in_(exp_ids))
            & (ExpRun.is_deleted == False)
            & (ExpRun.is_successful == True)
        )
    )
    for run in runs:
        run.exp = exps[run.exp_id]

    data = report.args.value.get_data(runs)
    return jsonify(
        {
            "type": report.args.type,
            "data": {
                "data": data.data,
                "xindex": [serialize_index(idx, exps) for idx in data.xindex],
                "yindex": [serialize_index(idx, exps) for idx in data.yindex],
            },
        }
    )


@report_bp.route(f"/{report_bp.name}/preview", methods=["POST"])
def preview_report():
    posted_record = request.json

    if "exps" not in posted_record:
        raise BadRequest("Missing `exps` field")
    exp_ids = exp_ids_deser(posted_record["exps"])
    if len(exp_ids) == 0:
        raise BadRequest("Empty `exps` field")
    exps = {e.id: e for e in Exp.select(Exp.id, Exp.name).where(Exp.id.in_(exp_ids))}

    raw_record = {}
    for name, field in Report._meta.fields.items():
        if name in posted_record and name != "id":
            try:
                raw_record[name] = report_deserializers[name](posted_record[name])
            except ValueError as e:
                raise ValueError(f"Field `{name}` {str(e)}")
    report = Report(**raw_record)

    # gather runs of experiments
    runs = list(
        ExpRun.select(
            ExpRun.id, ExpRun.params, ExpRun.aggregated_primitive_outputs, ExpRun.exp
        ).where(
            (ExpRun.exp.in_(exp_ids))
            & (ExpRun.is_deleted == False)
            & (ExpRun.is_successful == True)
        )
    )
    for run in runs:
        run.exp = exps[run.exp_id]

    data = report.args.value.get_data(runs)
    return jsonify(
        {
            "type": report.args.type,
            "data": {
                "data": data.data,
                "xindex": [serialize_index(idx, exps) for idx in data.xindex],
                "yindex": [serialize_index(idx, exps) for idx in data.yindex],
            },
        }
    )


def handle_internal_error(error):
    return jsonify({"message": str(error)}), 400


class ExperimentNotFound(Exception):
    pass


report_bp.register_error_handler(InvalidIndexError, handle_internal_error)
report_bp.register_error_handler(ExperimentNotFound, handle_internal_error)


def serialize_index(index: Index, exps: dict[int, Exp]):
    if index.attr[0] == EXP_INDEX_FIELD:
        if len(unk_exps := set(index.children.keys()).difference(exps.keys())) > 0:
            raise ExperimentNotFound(f"Experiments not found. Id: {list(unk_exps)}")

        return {
            "attr": ("Experiment",),
            "children": [
                (exps[eid].name, [serialize_index(c, exps) for c in lst])  # type: ignore
                for eid, lst in index.children.items()
            ],
        }
    return {
        "attr": index.attr,
        "children": [
            (v, [serialize_index(c, exps) for c in lst])
            for v, lst in index.children.items()
        ],
    }



from copy import deepcopy
from pathlib import Path
import orjson
from typing import List, Union, Dict, get_type_hints, Type, Any
from dataclasses import is_dataclass, asdict, dataclass, fields, Field, replace

from osin.types.pyobject_type import PyObjectType

DataClassInstance = Any


def are_valid_parameters(
    params: Union[
        DataClassInstance, List[DataClassInstance], Dict[str, DataClassInstance]
    ]
):
    """Check if the parameters are valid"""
    if isinstance(params, list):
        return all(is_dataclass(param) for param in params)
    elif isinstance(params, dict):
        return all(
            isinstance(name, str) and is_dataclass(param)
            for name, param in params.items()
        )
    else:
        assert is_dataclass(params), "Parameters must be an instance of a dataclass"


def param_as_dict(param: DataClassInstance) -> dict:
    """Convert a dataclass to a dictionary"""
    if not is_dataclass(param):
        raise TypeError("Parameter must be an instance of a dataclass")

    if hasattr(param, "to_dict"):
        return param.to_dict()
    return asdict(param)



from osin.types.pyobject_type import (
    PyObjectType,
    PRIMITIVE_TYPES,
    Number,
)
from osin.types.pyobject import PyObject, OTable, OImage, OAudio
from osin.types.primitive_type import NestedPrimitiveOutput, NestedPrimitiveOutputSchema
from osin.types.param_schema import ParamSchema



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Union, get_args

Number = Union[int, float]
NoneType = type(None)
TYPE_ALIASES = {"typing.List": "list", "typing.Dict": "dict", "typing.Set": "set"}
INSTANCE_OF = {
    "str": lambda ptype, x: isinstance(x, str),
    "int": lambda ptype, x: isinstance(x, int),
    "float": lambda ptype, x: isinstance(x, float),
    "typing.Union": lambda ptype, x: any(arg.is_instance(x) for arg in ptype.args),
}
PRIMITIVE_TYPES = {}


@dataclass
class PyObjectType:
    path: str
    args: List[PyObjectType]

    @staticmethod
    def from_tuple(t):
        return PyObjectType(path=t[0], args=[PyObjectType.from_tuple(x) for x in t[1]])

    @staticmethod
    def from_type_hint(hint) -> PyObjectType:
        global TYPE_ALIASES

        if hint.__module__ == "builtins":
            return PyObjectType(path=hint.__qualname__, args=[])

        if hasattr(hint, "__qualname__"):
            path = hint.__module__ + "." + hint.__qualname__
        else:
            # typically a class from the typing module
            if hasattr(hint, "_name") and hint._name is not None:
                path = hint.__module__ + "." + hint._name
            elif hasattr(hint, "__origin__") and hasattr(hint.__origin__, "_name"):
                # found one case which is typing.Union
                path = hint.__module__ + "." + hint.__origin__._name
            else:
                raise NotImplementedError(hint)

        if path in TYPE_ALIASES:
            # do it here because in python 3.10 typing.Dict has __qualname__
            path = TYPE_ALIASES[path]

        if path != "typing.Literal":
            args = [PyObjectType.from_type_hint(arg) for arg in get_args(hint)]
        else:
            args = [
                PyObjectType(encode_literal_value(arg), []) for arg in get_args(hint)
            ]

        return PyObjectType(path, args=args)

    @staticmethod
    def get_classpath(hint) -> str:
        global TYPE_ALIASES

        if hasattr(hint, "__qualname__"):
            path = hint.__module__ + "." + hint.__qualname__
        elif hint.__module__ == "builtins":
            path = hint.__qualname__
        elif hasattr(hint, "_name") and hint._name is not None:
            # typically a class from the typing module
            path = hint.__module__ + "." + hint._name
        elif hasattr(hint, "__origin__") and hasattr(hint.__origin__, "_name"):
            # found one case which is typing.Union
            path = hint.__module__ + "." + hint.__origin__._name
        else:
            raise NotImplementedError(hint)
        if path in TYPE_ALIASES:
            path = TYPE_ALIASES[path]
        return path

    def is_instance(self, value: Any):
        global INSTANCE_OF
        return INSTANCE_OF[self.path](self, value)

    def make_optional(self):
        return PyObjectType(
            path="typing.Union", args=[self, PyObjectType.from_type_hint(NoneType)]
        )

    def is_optional(self):
        return (
            self.path == "typing.Union"
            and len(self.args) == 2
            and (self.args[0].path == "NoneType" or self.args[1].path == "NoneType")
        )

    def __repr__(self) -> str:
        if self.path.startswith("typing."):
            path = self.path[7:]
        else:
            path = self.path

        if len(self.args) > 0:
            return f"{path}[{', '.join(repr(arg) for arg in self.args)}]"
        else:
            return path


for type_ in [str, int, bool, float, Number]:
    PRIMITIVE_TYPES[type_] = PyObjectType.from_type_hint(type_)


def encode_literal_value(value):
    if isinstance(value, str):
        return f"osin.types.str[{value}]"
    elif isinstance(value, int):
        return f"osin.types.int[{value}]"
    elif isinstance(value, bool):
        return f"osin.types.bool[{value}]"
    raise ValueError(f"Invalid value {value} for typing.Literal")



from __future__ import annotations
from dataclasses import MISSING, dataclass
from typing import Optional
from osin.models.report.index_schema import (
    AttrGetter,
    AttrValue,
    Attr,
    Index,
    IndexElement,
    IndexSchema,
    Record,
)


@dataclass
class BaseReport:
    x_axis: IndexSchema
    y_axis: IndexSchema
    z_values: list[tuple[Optional[int], list[AttrGetter]]]

    def get_data(self, records: list[Record]):
        """Get the report data"""
        xindex = self.x_axis.build_index(records)
        yindex = self.y_axis.build_index(records)

        data = {}
        for record in records:
            xelem = self.x_axis.get_index_element(record)
            yelem = self.y_axis.get_index_element(record)

            if xelem is None or yelem is None:
                continue

            if xelem not in data:
                data[xelem] = {}
            if yelem not in data[xelem]:
                data[xelem][yelem] = {}

            data_items = data[xelem][yelem]
            for expid, zvalues in self.z_values:
                if expid is not None and record.exp_id != expid:
                    continue
                for zvalue in zvalues:
                    zval = zvalue.get_value(record)
                    if zval is MISSING:
                        continue
                    key = zvalue.path
                    if key not in data_items:
                        data_items[key] = []
                    data_items[key].append(
                        {
                            "value": zval,
                            "record_id": record.id,
                        }
                    )

        datapoints = []
        for xitem, yitems_ in data.items():
            for yitem, zitems in yitems_.items():
                for zitem, values in zitems.items():
                    for value in values:
                        datapoints.append(
                            ReportDataPoint(
                                x=xitem,
                                y=yitem,
                                z=zitem,
                                record_id=value["record_id"],
                                record_value=value["value"],
                            )
                        )
        return ReportData(datapoints, xindex=xindex, yindex=yindex)

    @staticmethod
    def from_tuple(obj):
        return BaseReport(
            IndexSchema.from_tuple(obj[0]),
            IndexSchema.from_tuple(obj[1]),
            [
                (
                    expid if expid is not None else None,
                    [AttrGetter.from_tuple(o) for o in lst],
                )
                for expid, lst in obj[2]
            ],
        )

    def to_tuple(self):
        return (
            self.x_axis.to_tuple(),
            self.y_axis.to_tuple(),
            [
                (expid, [zval.to_tuple() for zval in zvals])
                for expid, zvals in self.z_values
            ],
        )


@dataclass
class ReportData:
    data: list[ReportDataPoint]
    xindex: list[Index]
    yindex: list[Index]


@dataclass
class ReportDataPoint:
    x: IndexElement
    y: IndexElement
    z: Attr
    record_id: int
    record_value: AttrValue



from __future__ import annotations

from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import certifi
import requests
from gena.deserializer import get_deserializer_from_type
from loguru import logger
from osin.apis.osin import Osin
from osin.apis.remote_exp import RemoteExpRun
from osin.models.exp import Exp, ExpRun
from osin.repository import OsinRepository
from osin.types import NestedPrimitiveOutputSchema, ParamSchema

ParamSchema_deser = get_deserializer_from_type(ParamSchema, {})
NestedPrimitiveOutputSchema_deser = get_deserializer_from_type(
    NestedPrimitiveOutputSchema, {}
)


class RemoteOsin(Osin):
    def __init__(self, endpoint: str, tmpdir: Path | str):
        super().__init__(tmpdir)
        self.endpoint = endpoint
        if self.endpoint.endswith("/"):
            self.endpoint = self.endpoint[:-1]
        self.tmpdir = Path(tmpdir)

    def _get(self, url: str, params: dict) -> dict:
        resp = requests.get(f"{self.endpoint}{url}", params=params)
        try:
            assert resp.status_code == 200
        except:
            logger.error(resp.text)
            raise
        return resp.json()

    def _post(self, url: str, data: dict) -> dict:
        resp = requests.post(f"{self.endpoint}{url}", json=data)
        try:
            assert resp.status_code == 200
        except:
            logger.error(resp.text)
            raise
        return resp.json()

    def _put(self, url: str, data: dict) -> dict:
        resp = requests.put(f"{self.endpoint}{url}", json=data)
        try:
            assert resp.status_code == 200
        except:
            logger.error(resp.text)
            raise
        return resp.json()

    def _cleanup(self, exp_run: RemoteExpRun):
        if exp_run.id not in self.cleanup_records:
            logger.debug("Cleaning up exp run: {}", exp_run.id)
            if exp_run.finished_time is None:
                # the user may forget to call finish_exp_run
                # we decide that it is still a failure
                try:
                    self.finish_exp_run(exp_run, is_successful=False)
                except:
                    finished_time = datetime.utcnow()
                    ExpRun.update(
                        is_finished=True,
                        is_successful=False,
                        finished_time=finished_time,
                    ).where(
                        ExpRun.id == exp_run.id
                    ).execute()  # type: ignore

                    raise
            else:
                finished_time = datetime.utcnow()
                ExpRun.update(
                    is_finished=True, is_successful=False, finished_time=finished_time
                ).where(
                    ExpRun.id == exp_run.id
                ).execute()  # type: ignore

    def _find_latest_exp(self, name: str) -> Optional[Exp]:
        exps = self._get(
            "/api/exp",
            {
                "name": name,
                "sorted_by": "-version",
                "limit": 1,
            },
        )["items"]
        if len(exps) == 0:
            return None
        else:
            exp = exps[0]
            return Exp(
                id=exp["id"],
                name=exp["name"],
                version=exp["version"],
                description=exp["description"],
                program=exp["program"],
                params=[ParamSchema_deser(p) for p in exp["params"]],
                aggregated_primitive_outputs=NestedPrimitiveOutputSchema(
                    exp["aggregated_primitive_outputs"]
                )
                if exp["aggregated_primitive_outputs"] is not None
                else None,
            )

    def _create_exp(self, exp: Exp) -> Exp:
        if exp.description is None or exp.params is None:
            raise ValueError(
                "Cannot create a new experiment without description and params"
            )
        obj = self._post(
            "/api/exp",
            {
                "name": exp.name,
                "version": exp.version,
                "description": exp.description,
                "program": exp.program,
                "params": [asdict(p) for p in exp.params],
                "aggregated_primitive_outputs": asdict(exp.aggregated_primitive_outputs)
                if exp.aggregated_primitive_outputs is not None
                else None,
            },
        )
        exp.id = obj["id"]
        return exp

    def _update_exp(self, exp_id: int, exp: Exp, fields: List[str]):
        data = {field: getattr(exp, field) for field in fields}
        if "params" in fields:
            data["params"] = [asdict(p) for p in exp.params]
        if (
            "aggregated_primitive_outputs" in fields
            and exp.aggregated_primitive_outputs is not None
        ):
            data["aggregated_primitive_outputs"] = asdict(
                exp.aggregated_primitive_outputs
            )
        self._put(f"/api/exp/{exp_id}", data)

    def _create_exprun(self, exprun: ExpRun) -> ExpRun:
        obj = self._post(
            "/api/exprun",
            {
                "exp_id": exprun.exp_id,
                "is_deleted": exprun.is_deleted,
                "is_finished": exprun.is_finished,
                "is_successful": exprun.is_successful,
                "has_invalid_agg_output_schema": exprun.has_invalid_agg_output_schema,
                "created_time": exprun.created_time.isoformat(),
                "finished_time": exprun.finished_time.isoformat()
                if exprun.finished_time is not None
                else None,
                "params": exprun.params,
                "metadata": asdict(exprun.metadata)
                if exprun.metadata is not None
                else None,
                "aggregated_primitive_outputs": exprun.aggregated_primitive_outputs,
            },
        )
        exprun.id = obj["id"]
        return exprun

    def _update_exprun(self, exprun_id: int, exprun: ExpRun, fields: List[str]):
        data = {field: getattr(exprun, field) for field in fields}
        if "metadata" in fields and exprun.metadata is not None:
            data["metadata"] = asdict(exprun.metadata)
        if "created_time" in fields:
            data["created_time"] = exprun.created_time.isoformat()
        if "finished_time" in fields and exprun.finished_time is not None:
            data["finished_time"] = exprun.finished_time.isoformat()

        self._put(
            f"/api/exprun/{exprun_id}",
            data,
        )

    def _upload_exprun(self, exprun: RemoteExpRun):
        files = {}
        for file in exprun.rundir.iterdir():
            if file.suffix in OsinRepository.ALLOWED_EXTENSIONS:
                files[file.stem] = (file.name, file.read_bytes())

        resp = requests.post(
            f"{self.endpoint}/api/exprun/{exprun.id}/upload", files=files
        )
        try:
            assert resp.status_code == 200
        except:
            logger.error(resp.text)
            raise



from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
import os
import socket
from peewee import (
    CharField,
    ForeignKeyField,
    TextField,
    IntegerField,
    BooleanField,
    DateTimeField,
    AutoField,
)
from playhouse.sqlite_ext import JSONField
import psutil
from osin.models.base import BaseModel
from osin.misc import json_dumps
from gena.custom_fields import DataClassField, ListDataClassField
from osin.types import (
    NestedPrimitiveOutput,
    NestedPrimitiveOutputSchema,
    ParamSchema,
)


@dataclass
class RunMetadata:
    hostname: str
    n_cpus: int
    # in bytes
    memory_usage: int

    @staticmethod
    def auto():
        return RunMetadata(
            hostname=socket.gethostname(),
            n_cpus=psutil.cpu_count(),
            memory_usage=psutil.Process(os.getpid()).memory_info().rss,
        )

    def to_dict(self):
        return {
            "hostname": self.hostname,
            "n_cpus": self.n_cpus,
            "memory_usage": self.memory_usage,
        }


class Exp(BaseModel):
    class Meta:
        indexes = (
            (("name", "version"), True),
            (("name",), False),
        )

    id: int = AutoField()  # type: ignore
    name: str = CharField(null=False)  # type: ignore
    version: int = IntegerField(null=False)  # type: ignore
    description: str = TextField()  # type: ignore
    program: str = TextField()  # type: ignore
    params: list[ParamSchema] = ListDataClassField(ParamSchema)  # type: ignore
    aggregated_primitive_outputs: NestedPrimitiveOutputSchema = DataClassField(NestedPrimitiveOutputSchema, null=True)  # type: ignore


class ExpRun(BaseModel):
    id: int = AutoField()  # type: ignore
    exp: Exp = ForeignKeyField(Exp, backref="runs", on_delete="CASCADE")  # type: ignore
    is_deleted: bool = BooleanField(default=False, index=True)  # type: ignore
    is_finished: bool = BooleanField(default=False, index=True)  # type: ignore
    is_successful: bool = BooleanField(default=False, index=True)  # type: ignore
    # whether the schema of the outputs is not consistent with the experiment.
    has_invalid_agg_output_schema: bool = BooleanField(default=False, index=True)  # type: ignore
    created_time: datetime = DateTimeField(default=datetime.utcnow)  # type: ignore
    finished_time: datetime = DateTimeField(null=True)  # type: ignore
    params: dict = JSONField(default={}, json_dumps=json_dumps)  # type: ignore
    metadata: RunMetadata = DataClassField(RunMetadata, null=True)  # type: ignore
    aggregated_primitive_outputs: NestedPrimitiveOutput = JSONField(default={}, json_dumps=json_dumps)  # type: ignore

    exp_id: int



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from osin.params_helper import DataClassInstance
from osin.models import ExpRunData
from osin.types import (
    NestedPrimitiveOutput,
    ParamSchema,
    NestedPrimitiveOutputSchema,
    PyObject,
    PyObjectType,
)

if TYPE_CHECKING:
    from osin.apis.osin import Osin


@dataclass
class RemoteExp:
    id: int
    name: str
    version: int
    params: List[ParamSchema]
    aggregated_primitive_outputs: Optional[NestedPrimitiveOutputSchema]
    osin: Osin

    def new_exp_run(
        self, params: Union[DataClassInstance, List[DataClassInstance]]
    ) -> RemoteExpRun:
        return self.osin.new_exp_run(self, params)


@dataclass
class RemoteExpRun:
    # id to the experiment run in osin
    id: int
    # id to the experiment in osin
    exp: RemoteExp
    created_time: datetime
    finished_time: Optional[datetime]
    rundir: Path
    osin: Osin
    pending_output: ExpRunData = field(default_factory=ExpRunData)

    def update_output(
        self,
        primitive: Optional[NestedPrimitiveOutput] = None,
        complex: Optional[Dict[str, PyObject]] = None,
    ):
        self.osin.update_exp_run_output(self, primitive, complex)
        return self

    def update_example_output(
        self,
        example_id: str,
        example_name: str = "",
        primitive: Optional[NestedPrimitiveOutput] = None,
        complex: Optional[Dict[str, PyObject]] = None,
    ):
        self.osin.update_example_output(
            self, example_id, example_name, primitive=primitive, complex=complex
        )
        return self

    def finish(self):
        self.osin.finish_exp_run(self)
        return self



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['osin',
 'osin.apis',
 'osin.controllers',
 'osin.formats',
 'osin.integrations',
 'osin.models',
 'osin.models.migration',
 'osin.models.report',
 'osin.types']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.2.2,<3.0.0',
 'gena>=1.6.3,<2.0.0',
 'h5py>=3.7.0,<4.0.0',
 'loguru>=0.6.0,<0.7.0',
 'orjson>=3.8.0,<4.0.0',
 'peewee>=3.15.2,<4.0.0',
 'psutil>=5.9.2,<6.0.0',
 'python-slugify>=6.1.2,<7.0.0',
 'ream2>=1.5.4,<2.0.0',
 't2-yada>=1.0.4,<2.0.0',
 'tornado>=6.2,<7.0']

setup_kwargs = {
    'name': 'osin',
    'version': '1.8.6',
    'description': 'Research and Experiments',
    'long_description': "# osin &middot; [![PyPI](https://img.shields.io/pypi/v/osin)](https://pypi.org/project/osin)\n\nThere are existing systems (e.g., neptune.ai, sacred) helping you organize, log data of your experiments. However, typically, the tasks of running the experiments are your responsible to bear. If you update your code and need to re-run your experiments, you may want to delete previous runs, which would be painful to have to do manually many times.\n\nWe rethink the experimenting process. Why don't we start with specifying the designed report (e.g., charts) and how to run/query to get the numbers to fill the report? This would free ones from manually starting/running the experiments and managing the experiment data. `osin` is a tool that helps you to achieve that goal.\n\nNote: this tool is expected to use locally or inside VPN network as it doesn't provide any protection against attackers.\n\n## Quick start\n\nStart the application:\n\n```bash\nDBFILE=%PATH_TO_DBFILE% python -m osin.main\n```\n\nOr start the services manually:\n\n```bash\nexport DBFILE=%PATH_TO_DBFILE%\npython -m osin.worker # start worker to run jobs\npython -m osin.server # start the server so clients can send job result\nstreamlit run osin/ui/dashboard.py # start a dashboard to view/create reports\n```\n\nYou will start by designing the output that your experiments will produce. For example:\n\n```yaml\n\n```\n",
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
import atexit
from datetime import datetime
from pathlib import Path
import shutil
from typing import Dict, List, Optional, Set, Union
from loguru import logger
import orjson

from osin.apis.remote_exp import RemoteExp, RemoteExpRun
from osin.misc import get_caller_python_script, orjson_dumps
from osin.models.exp import Exp, ExpRun, NestedPrimitiveOutput, RunMetadata
from osin.repository import OsinRepository
from osin.types import NestedPrimitiveOutputSchema, PyObject

from osin.params_helper import DataClassInstance, param_as_dict
from osin.types.param_schema import ParamSchema
from osin.types.primitive_type import validate_primitive_data
from osin.models.exp_data import ExampleData, Record


class Osin(ABC):
    """This class provides two methods to communicate with Osin, either locally or remotely via http protocol"""

    def __init__(self, osin_dir: Union[Path, str]):
        self.osin_keeper = OsinRepository(osin_dir)
        self.cleanup_records: Set[int] = set()

    @staticmethod
    def local(osin_dir: Union[Path, str]):
        from osin.apis.local_osin import LocalOsin

        return LocalOsin(osin_dir)

    @staticmethod
    def remote(endpoint: str, tmpdir: Union[Path, str] = "/tmp/osin"):
        from osin.apis.remote_osin import RemoteOsin

        return RemoteOsin(endpoint, tmpdir)

    def init_exp(
        self,
        name: str,
        version: int,
        description: Optional[str] = None,
        program: Optional[str] = None,
        params: Optional[Union[DataClassInstance, List[DataClassInstance]]] = None,
        aggregated_primitive_outputs: Optional[NestedPrimitiveOutputSchema] = None,
        update_param_schema: bool = False,
    ) -> RemoteExp:
        """Init an experiment in Osin.

        If the experiment already exists, the input version must be the latest one, and the parameters must match.
        If the experiment does not exist, it will be created with the given parameters.

        Args:
            name: Name of the experiment
            version: Version of the experiment
            description: Description of the experiment
            program: The python program to invoke the experiment
            params: The parameters of the experiment.
            aggregated_primitive_outputs: The aggregated outputs of the experiment.
                If not provided, it will be inferred automatically when the experiment is run.
        """
        exp = self._find_latest_exp(name)
        if exp is None:
            exp = self._create_exp(
                Exp(
                    name=name,
                    description=description,
                    version=version,
                    program=program or get_caller_python_script(),
                    params=[
                        ParamSchema.get_schema(p)
                        for p in (params if isinstance(params, list) else [params])
                    ],
                    aggregated_primitive_outputs=aggregated_primitive_outputs,
                )
            )
        else:
            if exp.version > version:
                raise ValueError("Cannot create an older version of an experiment")
            elif exp.version == version:
                if update_param_schema:
                    exp.params = [
                        ParamSchema.get_schema(p)
                        for p in (params if isinstance(params, list) else [params])
                    ]
                    self._update_exp(exp.id, exp, ["params"])
            else:
                exp = self._create_exp(
                    Exp(
                        name=name,
                        description=description,
                        version=version,
                        program=program or get_caller_python_script(),
                        params=[
                            ParamSchema.get_schema(p)
                            for p in (params if isinstance(params, list) else [params])
                        ],
                        aggregated_primitive_outputs=aggregated_primitive_outputs,
                    )
                )

        return RemoteExp(
            id=exp.id,
            name=exp.name,
            version=exp.version,
            params=exp.params,
            aggregated_primitive_outputs=exp.aggregated_primitive_outputs,
            osin=self,
        )

    def new_exp_run(
        self, exp: RemoteExp, params: Union[DataClassInstance, List[DataClassInstance]]
    ) -> RemoteExpRun:
        """Create a new run for an experiment."""
        output = {}
        for param in params if isinstance(params, list) else [params]:
            output.update(param_as_dict(param))
        exp_run = self._create_exprun(ExpRun(exp=exp.id, params=output))

        rundir = self.osin_keeper.get_exp_run_dir(exp, exp_run)
        if rundir.exists():
            shutil.rmtree(rundir)
        rundir.mkdir(parents=True)

        with open(rundir / "params.json", "wb") as f:
            f.write(orjson_dumps(output, option=orjson.OPT_INDENT_2))

        remote_exp_run = RemoteExpRun(
            id=exp_run.id,
            exp=exp,
            rundir=rundir,
            created_time=exp_run.created_time,
            finished_time=None,
            osin=self,
        )

        atexit.register(
            self._cleanup,
            exp_run=remote_exp_run,
        )
        return remote_exp_run

    def finish_exp_run(self, exp_run: RemoteExpRun, is_successful: bool = True):
        """Flush whatever remaining in experiment run that haven't sent to the server before stopping the experiment run."""
        exp_run.finished_time = datetime.utcnow()

        self.osin_keeper.get_exp_run_data_format(exp_run.exp, exp_run).save_run_data(
            exp_run.pending_output,
            self.osin_keeper.get_exp_run_data_file(exp_run.exp, exp_run),
        )

        metadata = RunMetadata.auto()
        # save metadata
        self.osin_keeper.get_exp_run_metadata_file(exp_run.exp, exp_run).write_bytes(
            orjson_dumps(
                {
                    "created_time": exp_run.created_time.isoformat(),
                    "finished_time": exp_run.finished_time.isoformat(),
                    "duration": (
                        exp_run.finished_time - exp_run.created_time
                    ).total_seconds(),
                    **metadata.to_dict(),
                },
                option=orjson.OPT_INDENT_2,
            )
        )
        # check if agg_lit_outputs matches with the schema.
        if exp_run.exp.aggregated_primitive_outputs is None:
            exp_run.exp.aggregated_primitive_outputs = (
                NestedPrimitiveOutputSchema.infer_from_data(
                    exp_run.pending_output.aggregated.primitive
                )
            )
            has_invalid_agg_output_schema = False
            self._update_exp(
                exp_run.exp.id,
                Exp(
                    aggregated_primitive_outputs=exp_run.exp.aggregated_primitive_outputs
                ),
                ["aggregated_primitive_outputs"],
            )
        else:
            has_invalid_agg_output_schema = (
                not exp_run.exp.aggregated_primitive_outputs.does_data_match(
                    exp_run.pending_output.aggregated.primitive
                )
            )

        self.osin_keeper.get_exp_run_success_file(exp_run.exp, exp_run).touch()
        self._upload_exprun(exp_run)
        self._update_exprun(
            exp_run.id,
            ExpRun(
                is_finished=True,
                is_successful=is_successful,
                finished_time=exp_run.finished_time,
                has_invalid_agg_output_schema=has_invalid_agg_output_schema,
                metadata=metadata,
                aggregated_primitive_outputs=exp_run.pending_output.aggregated.primitive,
            ),
            [
                "is_finished",
                "is_successful",
                "finished_time",
                "has_invalid_agg_output_schema",
                "metadata",
                "aggregated_primitive_outputs",
            ],
        )
        self.cleanup_records.add(exp_run.id)

    def update_exp_run_output(
        self,
        exp_run: RemoteExpRun,
        primitive: Optional[NestedPrimitiveOutput] = None,
        complex: Optional[Dict[str, PyObject]] = None,
    ):
        if primitive is not None:
            validate_primitive_data(primitive)
            exp_run.pending_output.aggregated.primitive.update(primitive)
        if complex is not None:
            exp_run.pending_output.aggregated.complex.update(complex)

    def update_example_output(
        self,
        exp_run: RemoteExpRun,
        example_id: str,
        example_name: str = "",
        primitive: Optional[NestedPrimitiveOutput] = None,
        complex: Optional[Dict[str, PyObject]] = None,
    ):
        if primitive is not None:
            validate_primitive_data(primitive)

        if example_id in exp_run.pending_output.individual:
            exp_run.pending_output.individual[example_id].name = example_name
            exp_run.pending_output.individual[example_id].data.primitive.update(
                primitive or {}
            )
            exp_run.pending_output.individual[example_id].data.complex.update(
                complex or {}
            )
        else:
            exp_run.pending_output.individual[example_id] = ExampleData(
                id=example_id,
                name=example_name,
                data=Record(
                    primitive=primitive or {},
                    complex=complex or {},
                ),
            )

    def _cleanup(self, exp_run: RemoteExpRun):
        if exp_run.id not in self.cleanup_records:
            logger.debug("Cleaning up exp run: {}", exp_run.id)
            if exp_run.finished_time is None:
                # the user may forget to call finish_exp_run
                # we decide that it is still a failure
                try:
                    self.finish_exp_run(exp_run, is_successful=False)
                except:
                    finished_time = datetime.utcnow()
                    self._update_exprun(
                        exp_run.id,
                        ExpRun(
                            is_finished=True,
                            is_successful=False,
                            finished_time=finished_time,
                        ),
                        ["is_finished", "is_successful", "finished_time"],
                    )
                    raise
            else:
                finished_time = datetime.utcnow()
                self._update_exprun(
                    exp_run.id,
                    ExpRun(
                        is_finished=True,
                        is_successful=False,
                        finished_time=finished_time,
                    ),
                    ["is_finished", "is_successful", "finished_time"],
                )

    @abstractmethod
    def _find_latest_exp(self, name: str) -> Optional[Exp]:
        pass

    @abstractmethod
    def _create_exp(self, exp: Exp) -> Exp:
        pass

    @abstractmethod
    def _update_exp(self, exp_id: int, exp: Exp, fields: List[str]):
        pass

    @abstractmethod
    def _create_exprun(self, exprun: ExpRun) -> ExpRun:
        pass

    @abstractmethod
    def _update_exprun(self, exprun_id: int, exprun: ExpRun, fields: List[str]):
        pass

    @abstractmethod
    def _upload_exprun(self, exprun: RemoteExpRun):
        pass



from osin.misc import json_dumps
from osin.models.base import BaseModel
from osin.models.exp import Exp
from peewee import (
    ForeignKeyField,
    AutoField,
)
from playhouse.sqlite_ext import JSONField


class ExpRunView(BaseModel):
    id: int = AutoField()  # type: ignore
    exp = ForeignKeyField(Exp, on_delete="CASCADE")
    config: dict = JSONField(default={}, json_dumps=json_dumps)  # type: ignore



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict

from osin.types import NestedPrimitiveOutput, PyObject


@dataclass
class Record:
    primitive: NestedPrimitiveOutput = field(default_factory=dict)
    complex: Dict[str, PyObject] = field(default_factory=dict)

    def to_dict(self):
        return {
            "primitive": self.primitive,
            "complex": {k: v.to_dict() for k, v in self.complex.items()},
        }


@dataclass
class ExampleData:
    id: str
    name: str
    data: Record = field(default_factory=Record)

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "data": self.data.to_dict(),
        }


@dataclass
class ExpRunData:
    aggregated: Record = field(default_factory=Record)
    individual: Dict[str, ExampleData] = field(default_factory=dict)

    def to_dict(self):
        return {
            "aggregated": self.aggregated.to_dict(),
            "individual": [v.to_dict() for k, v in self.individual.items()],
        }



from __future__ import annotations
from typing import Dict, Generic, Optional, List
import time
from ream.actors.base import BaseActor, E, P
from osin.apis.remote_exp import RemoteExp
from contextlib import contextmanager
from dataclasses import make_dataclass
from osin.apis.osin import Osin
from ream.params_helper import DataClassInstance


class OsinActor(Generic[E, P], BaseActor[E, P]):
    _osin: Optional[Osin] = None

    def __init__(self, params: P, dep_actors: Optional[List[BaseActor]] = None):
        super().__init__(params, dep_actors)
        self._exp: Optional[RemoteExp] = None

    @contextmanager
    def new_exp_run(self, **kwargs):
        """Start a new experiment run"""
        if self._osin is None:
            yield None
        else:
            exp_params = self.get_exp_run_params()
            if len(kwargs) > 0:
                C = make_dataclass(
                    "DynamicParams", [(k, type(v)) for k, v in kwargs.items()]
                )
                exp_params.append(C(**kwargs))

            if self._exp is None:
                self.logger.debug("Setup experiments...")
                cls = self.__class__
                assert cls.__doc__ is not None, "Please add docstring to the class"
                self._exp = self._osin.init_exp(
                    name=getattr(cls, "NAME", cls.__name__),  # type: ignore
                    version=getattr(cls, "EXP_VERSION", 1),
                    description=cls.__doc__,
                    params=exp_params,
                )

            exprun = self._exp.new_exp_run(exp_params)
            yield exprun
            if exprun is not None:
                self.logger.debug(
                    "Flushing run data of the experiment {}", self._exp.name
                )
                start = time.time()
                exprun.finish()
                end = time.time()
                self.logger.debug(
                    "\tFlushing run data took {:.3f} seconds", end - start
                )

    def get_exp_run_params(self) -> List[DataClassInstance]:
        """Get the parameters of the experiment run"""
        # we can use id() to guarantee the uniqueness of the actor
        # but instead we use the class
        stack: List[BaseActor] = [self]
        params = []
        type2id = {}
        while len(stack) > 0:
            actor = stack.pop()
            if actor.__class__ in type2id:
                assert id(actor) == type2id[actor.__class__]
            else:
                type2id[actor.__class__] = id(actor)
                params.append(actor.params)
            stack.extend(actor.dep_actors)

        return params
